from ..utils import AttrDict, ContextStack as BaseContextStack
from flask import request, g, session
import datetime
from contextlib import contextmanager
import functools


class ContextStack(BaseContextStack):
    def pop(self):
        exported_vars = self.top.get_exported_vars()
        super(ContextStack, self).pop()
        if self.top:
            self.top.vars.update(exported_vars)


_context_stack = ContextStack()
current_context = _context_stack.make_proxy()


def has_action_context():
    return _context_stack.top is not None


@contextmanager
def ensure_context(**vars):
    """Ensures that a context is in the stack, creates one otherwise.
    """
    ctx = _context_stack.top
    stacked = False
    if not ctx:
        ctx = Context()
        stacked = True
        _context_stack.push(ctx)
    ctx.update(vars)
    try:
        yield ctx
    finally:
        if stacked:
            _context_stack.pop()


class ContextExitException(Exception):
    """Exception to stop the execution of an action and of the following ones
    """
    def __init__(self, result=None, trigger_action_group=None):
        super(ContextExitException, self).__init__()
        self.result = result
        self.trigger_action_group = trigger_action_group


class TriggerActionGroupException(Exception):
    """Exception to stop the execution of an action and trigger another action group
    Normal execution flow will resume after that.
    """
    def __init__(self, group):
        super(TriggerActionGroupException, self).__init__()
        self.group = group


class Context(object):
    """Holds the data for an execution context. The data attribute is used to
    hold data which actions want to pass to each others without exposing them
    in the view. The vars attribute holds data for the template.
    """
    def __init__(self, vars=None, data=None, **kwargs):
        self.vars = AttrDict(vars or {}, **kwargs)
        self.data = AttrDict(data or {})
        self.export_vars = []
        self.executed_actions = set()

    def __getitem__(self, name):
        return self.vars[name]

    def __setitem__(self, name, value):
        self.vars[name] = value

    def __contains__(self, name):
        return name in self.vars

    def update(self, vars):
        self.vars.update(vars)

    def get(self, name, default=None):
        return self.vars.get(name, default)

    def clone(self, **override_vars):
        """Creates a copy of this context"""
        c = Context(self.vars, self.data)
        c.executed_actions = set(self.executed_actions)
        c.vars.update(override_vars)
        return c

    def trigger_action_group(self, group):
        raise TriggerActionGroupException(group)

    def exit(self, result=None, trigger_action_group=None):
        raise ContextExitException(result, trigger_action_group)

    def get_exported_vars(self):
        return dict((k, self.vars[k]) for k in self.export_vars)

    def __enter__(self):
        _context_stack.push(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        _context_stack.pop()
        return False


def request_context(app, request):
    """Creates a Context instance from the given request object
    """
    vars = {}
    if request.view_args is not None:
        vars.update(request.view_args)
    vars.update({
        "request": request,
        "GET": AttrDict(request.args.to_dict()),
        "POST" : AttrDict(request.form.to_dict()),
        "app": app,
        "config": app.config,
        "session": session,
        "g": g,
        "now": datetime.datetime.now,
        "utcnow": datetime.datetime.utcnow,
        "today": datetime.date.today})
    context = Context(vars)
    context.vars["current_context"] = context
    return context


def pass_context_var(*vars):
    def decorator(f):
        @functools.wraps(f)
        def wrap(*args, **kwargs):
            kwargs.update(dict((k, current_context[k]) for k in vars))
            return f(*args, **kwargs)
        return wrap
    return decorator


import functools
from flask import request, abort, current_app, has_request_context
from .utils import unknown_value, FlagContextStack


def _value_from_multidict(v, aslist=False):
    if aslist:
        return v
    if len(v) == 0:
        return None
    if len(v) == 1:
        return v[0]
    return v


class MissingRequestParam(Exception):
    pass


def get_request_param_value(name, location=None, aslist=False):
    if (not location or location == 'view_args') and name in request.view_args:
        if aslist and not isinstance(request.view_args[name], list):
            return [request.view_args[name]]
        return request.view_args[name]

    if not request.is_json:
        if not location or location == 'values':
            if name in request.values:
                return _value_from_multidict(request.values.getlist(name), aslist)
        elif location == 'args' and name in request.args:
            return _value_from_multidict(request.args.getlist(name), aslist)
        elif location == 'form' and name in request.form:
            return _value_from_multidict(request.form.getlist(name), aslist)

    if (not location or location == 'json') and request.is_json:
        data = request.get_json(silent=True)
        if isinstance(data, dict) and name in data:
            if aslist and not isinstance(data[name], list):
                return [data[name]]
            return data[name]

    if (not location or location == 'files') and name in request.files:
        return _value_from_multidict(request.files.getlist(name), aslist)

    raise MissingRequestParam("Request parameter '%s' is missing" % name)


class RequestParam(object):
    def __init__(self, name, type=None, nullable=False, required=False, loader=None, validator=None,
                 dest=None, location=None, aslist=False, aslist_iter_loader=True, help=None,
                 default=unknown_value, **loader_kwargs):
        self.name = name
        self.type = type
        self.nullable = nullable
        self.required = required
        self.loader = loader
        self.loader_kwargs = loader_kwargs
        self.validator = validator
        self.dest = dest
        self.location = location
        self.aslist = aslist
        self.aslist_iter_loader = aslist_iter_loader
        self.help = help
        self.default = default

    @property
    def names(self):
        return self.name if isinstance(self.name, tuple) else (self.name,)

    @property
    def types(self):
        if not self.type:
            return
        return self.type if isinstance(self.type, tuple) else (self.type,)

    @property
    def names_types(self):
        names = self.names
        fallback_type = None
        types = []
        if self.type:
            types = self.types
            fallback_type = types[-1]
        return [(names[i], types[i] if i in types else fallback_type) for i in range(len(names))]

    @property
    def dests(self):
        if self.dest is None:
            return self.names
        elif self.dest and not isinstance(self.dest, tuple):
            return (self.dest,)
        return self.dest

    def process(self, request_data=None):
        try:
            values = self.extract(request_data)
        except MissingRequestParam:
            if self.required and self.default is unknown_value:
                current_app.logger.debug('Missing required parameter: %s' % ", ".join(self.names))
                abort(400)
            elif not self.default is unknown_value:
                if not isinstance(self.name, tuple):
                    values = [self.default]
                else:
                    values = self.default
            else:
                return dict()

        values = self.load(*values)
        if not self.validate(*values):
            current_app.logger.debug('Invalid parameter: %s' % ", ".join(self.names))
            abort(400)
        if self.dests is False:
            return {}
        if len(self.dests) != len(values):
            raise Exception('Mismatch between length of dest and number of values')
        return dict(zip(self.dests, values))

    def extract(self, request_data=None):
        values = []
        for name, type in self.names_types:
            if request_data:
                if name not in request_data:
                    raise MissingRequestParam("Request parameter '%s' is missing" % name)
                value = request_data[name]
                if self.aslist and not isinstance(value, (tuple, list)):
                    value = [value]
            else:
                value = get_request_param_value(name, self.location, self.aslist)
            if value is None and not self.nullable:
                raise MissingRequestParam("Request parameter '%s' is missing" % name)
            if type and value is not None:
                if self.aslist:
                    value = map(lambda v: self.coerce(v, type), value)
                else:
                    value = self.coerce(value, type)
            values.append(value)
        return values

    def coerce(self, value, type):
        try:
            if type is bool:
                if isinstance(value, (str, unicode)):
                    return value.lower() not in ('False', 'false', 'no', '0')
                return bool(value)
            return type(value)
        except Exception as e:
            current_app.log_exception(e)
            abort(400)

    def load(self, *values):
        if self.loader:
            if self.aslist and self.aslist_iter_loader:
                rv = []
                for t in zip(*values):
                    rv.append(self.loader(*t, **self.loader_kwargs))
                return (rv,)
            else:
                rv = self.loader(*values, **self.loader_kwargs)
                if not isinstance(rv, tuple):
                    return (rv,)
                return rv
        return values

    def validate(self, *values):
        return not self.validator or self.validator(*values)

    def update_kwargs(self, kwargs, request_data=None):
        for name in self.names:
            kwargs.pop(name, None)
        kwargs.update(self.process(request_data))
        return kwargs


auto_inject_request_params = FlagContextStack()


def update_kwargs_with_request_params(kwargs, request_params, request_data=None):
    for param in request_params:
        param.update_kwargs(kwargs, request_data)
    return kwargs


def wrap_request_param_func(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if auto_inject_request_params.active() and has_request_context():
            update_kwargs_with_request_params(kwargs, wrapper.request_params)
        return func(*args, **kwargs)
    wrapper.request_params = []
    return wrapper


def request_param(name, cls=RequestParam, append=True, **kwargs):
    def decorator(func):
        if not append or not hasattr(func, 'request_params'):
            wrapper = wrap_request_param_func(func)
        else:
            wrapper = func
        if isinstance(name, RequestParam):
            wrapper.request_params.append(name)
        else:
            wrapper.request_params.append(cls(name, **kwargs))
        return wrapper
    return decorator


def partial_request_param(name=None, cls=RequestParam, **kwargs):
    def decorator(name_=None, **kw):
        return request_param(name_ or name, cls=cls, **dict(kwargs, **kw))
    decorator.request_param = cls(name, **kwargs)
    return decorator


def request_params(params):
    def decorator(func):
        if isinstance(params, list):
            for i, param in enumerate(params):
                func = request_param(param)(func)
        elif params:
            for name, kwargs in params.iteritems():
                func = request_param(name, **kwargs)(func)
        return func
    decorator.params = params
    return decorator


def request_param_loader(name, **kwargs):
    def decorator(loader_func):
        return request_param(name, loader=loader_func, **kwargs)
    return decorator


def partial_request_param_loader(name=None, **kwargs):
    def decorator(loader_func):
        return partial_request_param(name, loader=loader_func, **kwargs)
    return decorator



from flask import Flask, request
from flask.config import ConfigAttribute, Config as FlaskConfig
from flask.ctx import AppContext, has_request_context
from flask.helpers import locked_cached_property
from jinja2 import ChoiceLoader, FileSystemLoader
from .utils import AttrDict, RequirementMissingError, JSONEncoder, deep_update_dict
from .actions import (ActionList, ActionRegistry, action_resolver, request_context,\
                     current_context, ContextExitException, common_actions, ensure_context)
from .views import ViewContainerMixin
from .decorators import register_hooks
from .templating import JinjaLoader, configure_environment
from .features import FeatureRegistry
from .request_params import auto_inject_request_params
from .marshaller import marshaller_context
import inspect
import os
import yaml
import errno
import sys


class Config(FlaskConfig):
    """Subclass of Flask's Config class to add support to load from YAML file
    """

    def from_json(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as json_file:
                obj = json.loads(json_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_yaml(self, filename, silent=False, deep_update=False):
        filename = os.path.join(self.root_path, filename)

        try:
            with open(filename) as yaml_file:
                obj = yaml.load(yaml_file.read())
        except IOError as e:
            if silent and e.errno in (errno.ENOENT, errno.EISDIR):
                return False
            e.strerror = 'Unable to load configuration file (%s)' % e.strerror
            raise
        return self.from_mapping(obj, _deep_update=deep_update)

    def from_mapping(self, *mapping, **kwargs):
        mappings = []
        if len(mapping) == 1:
            if hasattr(mapping[0], 'items'):
                mappings.append(mapping[0].items())
            else:
                mappings.append(mapping[0])
        elif len(mapping) > 1:
            raise TypeError(
                'expected at most 1 positional argument, got %d' % len(mapping)
            )
        deep_update = kwargs.pop('_deep_update', False)
        mappings.append(kwargs.items())
        for mapping in mappings:
            if deep_update:
                deep_update_dict(self, dict((k.upper(), v) for (k, v) in mapping))
            else:
                for (key, value) in mapping:
                    self[key.upper()] = value
        return True

    def from_file(self, filename, silent=False, deep_update=False):
        if filename.endswith(".py"):
            self.from_pyfile(filename, silent=silent)
        elif filename.endswith(".js") or filename.endswith(".json"):
            self.from_json(filename, silent=silent, deep_update=deep_update)
        elif filename.endswith(".yml"):
            self.from_yaml(filename, silent=silent, deep_update=deep_update)
        else:
            raise RuntimeError("Unknown config file extension")

    def resolve_includes(self, relative_to=".", key="INCLUDE_FILES"):
        for spec in self.pop(key, []):
            if not isinstance(spec, dict):
                spec = {"filename": spec}
            filename = os.path.join(relative_to, spec["filename"])
            self.from_file(filename, silent=spec.get("silent", False),
                deep_update=spec.get("deep_update", False))


class FrascoAppContext(AppContext):
    def push(self):
        action_resolver.push(self.app.actions)
        super(FrascoAppContext, self).push()

    def pop(self, exc=None):
        super(FrascoAppContext, self).pop(exc)
        action_resolver.pop()


class Frasco(ViewContainerMixin, Flask):
    """Flask subclass that adds support to register feature and actions
    """
    config_class = Config
    json_encoder = JSONEncoder

    processes = ConfigAttribute('PROCESSES')
    services_url_prefix = ConfigAttribute('SERVICES_URL_PREFIX')
    services_subdomain = ConfigAttribute('SERVICES_SUBDOMAIN')

    def __init__(self, import_name, **kwargs):
        self.feature_blueprint_names = set()
        with_common_actions = kwargs.pop('with_common_actions', True)
        super(Frasco, self).__init__(import_name, **kwargs)
        self.features = FeatureRegistry()
        self.actions = ActionRegistry()
        self.services = AttrDict()
        self.views = AttrDict()
        self.request_actions = ActionList()
        self.before_request(self.execute_before_request_actions)
        self.after_request(self.execute_after_request_actions)
        if with_common_actions:
            self.actions.register(common_actions)

    def make_config(self, instance_relative=False):
        root_path = self.root_path
        if instance_relative:
            root_path = self.instance_path
        return self.config_class(root_path, dict(self.default_config,
            PROCESSES=[],
            SERVICES_URL_PREFIX='/api',
            SERVICES_SUBDOMAIN=None))

    def update_template_context(self, context):
        context["current_context"] = current_context
        super(Frasco, self).update_template_context(context)

    def full_dispatch_request(self):
        with request_context(self, request), marshaller_context.once(), auto_inject_request_params.once():
            return super(Frasco, self).full_dispatch_request()

    def preprocess_request(self):
        try:
            return super(Frasco, self).preprocess_request()
        except ContextExitException as e:
            return e.result

    def process_response(self, response):
        # ensure_context() is needed because process_response() can be called outside
        # of full_dispatch_request in case of error
        with ensure_context():
            try:
                return super(Frasco, self).process_response(response)
            except ContextExitException as e:
                return self.make_response(e.result)

    def app_context(self):
        return FrascoAppContext(self)

    def log_exception(self, exc_info):
        if not isinstance(exc_info, tuple):
            exc_info = sys.exc_info()
        if has_request_context():
            super(Frasco, self).log_exception(exc_info)
        else:
            self.logger.error('Exception [outside of request context]', exc_info=exc_info)

    def create_jinja_environment(self):
        env = super(Frasco, self).create_jinja_environment()
        env.globals.update(app=self)
        configure_environment(env)
        macro_file = os.path.join(self.root_path, "macros.html")
        if os.path.exists(macro_file):
            env.macros.register_file(macro_file)
        macro_dir = os.path.join(self.root_path, "macros")
        if os.path.exists(macro_dir):
            env.macros.register_directory(macro_dir)
        return env

    def create_global_jinja_loader(self):
        return JinjaLoader(self)

    @locked_cached_property
    def jinja_loader(self):
        loader = ChoiceLoader([])
        if self.template_folder is not None:
            loader.loaders.append(FileSystemLoader(os.path.join(self.root_path,
                                                 self.template_folder)))
        return loader

    def register_blueprint(self, blueprint, **options):
        with self.app_context(), self.actions:
            return super(Frasco, self).register_blueprint(blueprint, **options)

    def register_feature(self, feature, with_blueprints=True):
        if feature.requires is not None:
            for r in feature.requires:
                if r not in self.features:
                    raise RequirementMissingError("Feature '%s' requires '%s'" % (feature.name, r))
        self.features.register(feature)
        self.actions.register_many(feature.actions, feature.name)
        feature.init_app(self)
        if feature.commands:
            for cmd in feature.commands:
                self.cli.add_command(cmd)
        if feature.hooks is not None:
            for func, hooks in feature.hooks:
                register_hooks(func, hooks, self)
        if with_blueprints:
            feature.init_blueprints(self)
        return feature

    def init_feature_blueprints(self):
        for feature in self.features:
            feature.init_blueprints(self)

    def execute_before_request_actions(self):
        self.request_actions.execute(limit_groups=('before', None))

    def execute_after_request_actions(self, response):
        current_context["response"] = response
        self.request_actions.execute(limit_groups='after')
        return response

    def add_view(self, view):
        if inspect.isclass(view):
            view = view(self_var=self)
        self.views[view.name] = view
        with self.app_context(), self.actions:
            view.register(self)

    def register_service(self, service):
        self.services[service.name] = service
        self.actions.register_many(service.actions, service.name)
        if service.exposed:
            self.register_service_blueprint(service.as_blueprint())
        return service

    def register_service_blueprint(self, blueprint):
        self.register_blueprint(blueprint,
            url_prefix=self.services_url_prefix,
            subdomain=self.services_subdomain)

    @property
    def action(self):
        return self.actions.action

    def service(self, cls):
        self.register_service(cls())
        return cls


import re
from .utils import import_string
import datetime


class MissingVarError(Exception):
    pass


class Node(object):
    pass


class DataNode(Node):
    """An AST node representing some string data
    """
    def __init__(self, data):
        self.data = data

    def eval(self, context):
        return self.data

    def __repr__(self):
        return "DataNode(%s)" % self.data


class VarNode(Node):
    """An AST node representing a variable from the context
    """
    def __init__(self, name):
        self.name = name

    def eval(self, context):
        return context.get(self.name)

    def __repr__(self):
        return "VarNode(%s)" % self.name


class GetAttrNode(Node):
    """An AST node representing accessing an attribute of the child node
    """
    def __init__(self, node, attr):
        self.node = node
        self.attr = attr

    def eval(self, context):
        obj = self.node.eval(context)
        if obj:
            return getattr(obj, self.attr, None)

    def __repr__(self):
        return "GetAttrNode(%s, %s)" % (repr(self.node), self.attr)


class GetItemNode(Node):
    """An AST node representing accessing an item of the child node
    """
    def __init__(self, node, item):
        self.node = node
        self.item = item

    def eval(self, context):
        obj = self.node.eval(context)
        item = self.item
        if isinstance(item, Node):
            item = item.eval(context)
        try:
            return obj[item]
        except KeyError:
            pass

    def __repr__(self):
        return "GetItemNode(%s, %s)" % (repr(self.node), self.item)


class CallNode(Node):
    """An AST node representing calling a method of the child node
    """
    def __init__(self, node, method=None):
        self.node = node
        self.method = method

    def eval(self, context):
        obj = self.node.eval(context)
        if self.method is None:
            return obj()
        if not hasattr(obj, self.method):
            raise MissingVarError("Unknwon method '%' in option value" % self.method)
        return getattr(obj, self.method)()

    def __repr__(self):
        return "CallNode(%s, %s)" % (repr(self.node), self.method)


class DefaultNode(Node):
    """An AST node representing a default value for a missing variable
    """
    def __init__(self, node, default):
        self.node = node
        self.default = default

    def eval(self, context):
        try:
            value = self.node.eval(context)
            if value is not None:
                return value
        except:
            pass
        if isinstance(self.default, Node):
            return self.default.eval(context)
        return self.default

    def __repr__(self):
        return "DefaultNode(%s, %s)" % (repr(self.node), repr(self.default))


class NotNode(Node):
    """An AST node representing "not"
    """
    def __init__(self, node):
        self.node = node

    def eval(self, context):
        return not self.node.eval(context)

    def __repr__(self):
        return "NotNode(%s)" % repr(self.node)


class AndNode(Node):
    """An AST node representing "&&"
    """
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def eval(self, context):
        return self.left.eval(context) and self.right.eval(context)

    def __repr__(self):
        return "AndNode(%s, %s)" % (repr(self.left), repr(self.right))


class OrNode(Node):
    """An AST node representing "||"
    """
    def __init__(self, left, right):
        self.left = left
        self.right = right

    def eval(self, context):
        return self.left.eval(context) or self.right.eval(context)

    def __repr__(self):
        return "OrNode(%s, %s)" % (repr(self.left), repr(self.right))


class EvalNode(Node):
    """An AST node representing an eval expression.
    These expressions can use import(module_name) which will return
    a module.
    """
    import_regexp = re.compile(r"(import\(([a-z.]+)\))")

    def __init__(self, expr, catch_errors=True):
        self.expr = expr
        self.catch_errors = catch_errors
        for import_expr, name in self.import_regexp.findall(self.expr):
            self.expr = self.expr.replace(import_expr, "__import_module('%s')" % name)

    def eval(self, context):
        ctx = dict(context, __import_module=lambda n: import_string(n), datetime=datetime)
        try:
            return eval(self.expr, globals(), ctx)
        except:
            if not self.catch_errors:
                raise
            return None

    def __repr__(self):
        return "EvalNode(%s)" % self.expr


class ExpressionSyntaxError(Exception):
    pass


class Parser(object):
    """Parses a string and generates an AST
    Variables can be used using the dollar sign in front of their name. Attributes
    and items can be accessed as usual. Item names do not need to be strings (ie. use [foo] instead of ["foo"]).
    It is also possible to eval python code enclosed in {{ }}.
    """
    vars_regexp = re.compile(r"(\!?(?<!\\)\$\{?([a-zA-Z0-9_]+)(\(\))*(((\.[a-zA-Z0-9_]+(\(\))*)*(\[\$*[a-zA-Z_.0-9]+\])*)*)\}?)")
    eval_regexp = re.compile(r"\{\{([^}]+)\}\}")
    getattr_regexp = re.compile(r"\.([a-zA-Z0-9_]+)")
    getitem_regexp = re.compile(r"\[(\$*[a-zA-Z_.0-9]+)\]")
    call_regexp = re.compile(r"\.([a-zA-Z0-9_]+)\(\)")
    default_regexp = re.compile(r"([a-zA-Z0-9]+)")
    operators_regexps = [(re.compile(r" && "), AndNode), (re.compile(r" \|\| "), OrNode)]
    special_chars = [('\$', '$'), ('\|\|', '||'), ('\&\&', '&&')]

    def __init__(self):
        self.regexps = [(self.eval_regexp, self.parse_eval), (self.vars_regexp, self.parse_var)]

    def compile(self, expr, force_expr=False):
        if isinstance(expr, list):
            return self.compile_list(expr, force_expr)
        if isinstance(expr, dict):
            return self.compile_dict(expr, force_expr)
        if not isinstance(expr, (str, unicode)):
            return expr

        ast = self.parse(expr)
        if not force_expr and len(ast) == 1 and isinstance(ast[0], DataNode):
            return expr
        return Expression(ast, expr)

    def compile_list(self, lst, force_expr=False):
        rv = []
        for item in lst:
            rv.append(self.compile(item, force_expr))
        return rv

    def compile_dict(self, dct, force_expr=False):
        rv = {}
        for k, v in dct.iteritems():
            rv[k] = self.compile(v, force_expr)
        return rv

    def parse(self, expr):
        self.expr = expr
        nodes = []
        while True:
            match = False
            for regexp, func in self.regexps:
                r = regexp.search(expr)
                if r:
                    before_expr = expr[:r.start()]
                    if before_expr != "":
                        nodes.append(DataNode(self.cleanup_special_chars(before_expr)))
                    node, expr = func(r, expr[r.end():])
                    nodes.append(node)
                    match = True
                    break
            if match:
                continue
            if expr != "":
                nodes.append(DataNode(self.cleanup_special_chars(expr)))
            break

        return nodes

    def cleanup_special_chars(self, expr):
        for a, b in self.special_chars:
            expr = expr.replace(a, b)
        return expr

    def parse_var(self, match, remaining=None):
        node = VarNode(match.group(2))
        if match.group(3) is not None:
            node = CallNode(node)
        access = match.group(4)
        if access != "":
            node = self.parse_accessors(access, node)
        if match.group(1).startswith("!"):
            node = NotNode(node)

        if remaining is None:
            return node

        if remaining.startswith("|"):
            remaining = remaining[1:]
            m = self.vars_regexp.match(remaining)
            if m:
                node = DefaultNode(node, self.parse_var(m))
                remaining = remaining[m.end():]
            else:
                m = self.default_regexp.match(remaining)
                if not m:
                    raise ExpressionSyntaxError("Missing default value after default operator in '%s'" % self.expr)
                node = DefaultNode(node, m.group(1))
                remaining = remaining[m.end():]

        for regexp, node_cls in self.operators_regexps:
            r = regexp.match(remaining)
            if r:
                remaining = remaining[r.end():]
                match = self.vars_regexp.match(remaining)
                if not match:
                    raise ExpressionSyntaxError("Operators can only be used between variables in '%s'" % self.expr)
                right_node, remaining = self.parse_var(match, remaining[match.end():])
                node = node_cls(node, right_node)

        return (node, remaining)

    def parse_accessors(self, expr, node):
        r = self.call_regexp.match(expr)
        if r:
            return self.parse_accessors(expr[r.end():], CallNode(node, r.group(1)))

        r = self.getattr_regexp.match(expr)
        if r:
            return self.parse_accessors(expr[r.end():], GetAttrNode(node, r.group(1)))

        r = self.getitem_regexp.match(expr)
        if r:
            item = r.group(1)
            if item.startswith("$"):
                item = self.parse_var(item)
            return self.parse_accessors(expr[r.end():], GetItemNode(node, item))

        return node

    def parse_eval(self, match, remaining):
        return (EvalNode(match.group(1).strip()), remaining)


class Expression(object):
    """Represents an option value which needs to be evald
    """
    def __init__(self, ast, expr=None):
        self.ast = ast
        self.expr = expr

    def eval(self, context, text_type=unicode):
        result = []
        for node in self.ast:
            result.append(node.eval(context))
        if len(result) == 1:
            return result[0]
        return text_type("").join(map(text_type, result))

    def __str__(self):
        return self.expr

    def __repr__(self):
        return "<Expression(%s)>" % self.expr


def compile_expr(expr):
    """Compiles an expression
    """
    return Parser().compile(expr)


def eval_expr(expr, context):
    """Recursively evaluates a compiled expression using the specified context.
    Dict instances can contain a "__kwargs" key which will be used to update the
    dict with its content
    """
    if isinstance(expr, list):
        rv = []
        for item in expr:
            rv.append(eval_expr(item, context))
        return rv
    if isinstance(expr, dict):
        rv = {}
        for k, v in expr.iteritems():
            rv[k] = eval_expr(v, context)
        kwargs = rv.pop("__kwargs", None)
        if kwargs:
            rv.update(kwargs)
        return rv
    if isinstance(expr, Expression):
        return expr.eval(context)
    return expr


from ..utils import RequirementMissingError, ClassBoundDecoratorMixin
from ..expression import eval_expr
from .context import ensure_context, current_context
import inspect


class OptionMissingError(Exception):
    pass


class InvalidOptionError(Exception):
    pass


class Action(object):
    """Override this class to create a custom action
    """
    # Name of your action (it is mandatory to override it)
    name = None
    # Default group
    group = None
    # Allowed http methods for this action
    methods = None
    # When the function returns something, what should be the variable name
    as_ = None
    # A list of action names required for this action
    requires = []
    # Default options
    defaults = {}
    # If option is a string, it will be transformed to a dict using
    # default_option as the key
    default_option = "default"

    def __init__(self, options=None, group=None, as_=None, **kwargs):
        self.options = dict(self.defaults)
        if options:
            if not isinstance(options, dict):
                options = dict([(self.default_option, options)])
            self.options.update(options)
        self.options.update(kwargs)
        if group:
            self.group = group
        if as_:
            self.as_ = as_

    def init_view(self, view):
        """Called when an action is added to a view
        """

    def is_valid(self, context):
        """Checks through the previous_actions iterable if required actions have
        been executed
        """
        if self.requires:
            for r in self.requires:
                if not r in context.executed_actions:
                    raise RequirementMissingError("Action '%s' requires '%s'" % (self.name, r))
        return True

    def execute(self):
        with ensure_context():
            rv = self._execute(eval_expr(self.options, current_context.vars))
            g = None
            if inspect.isgenerator(rv):
                g = rv
                rv = g.next()
            if self.as_:
                current_context.vars[self.as_] = rv
            if g:
                try:
                    g.next()
                except StopIteration:
                    pass
        return rv

    def _execute(self, options):
        """Unless you want to override the evaluation of options, you should
        override this method to define the logic of your action
        """
        raise NotImplementedError

    def __repr__(self):
        return "<Action %s%s>" % (self.name, " in %s" % self.group if self.group else "")


class ActionFunction(ClassBoundDecoratorMixin):
    """A decorator to create action classes from functions or class methods.
    """
    # kwargs will be used as attributes of the action class
    def __init__(self, func, name=None, init_view_func=None, **kwargs):
        self.__name__ = self.name = name or func.__name__
        self.unbound_func = func
        self.func = self._wrap_func(func)
        self.init_view_func = None
        if init_view_func:
            self.init_view(init_view_func)

        def execute(s):
            self.as_ = s.as_
            rv = Action.execute(s)
            del self.as_
            return rv

        def _execute(s, options):
            self.as_ = s.as_
            rv = self.func(**options)
            s.as_ = self.as_
            return rv

        def init_view(s, view):
            if self.init_view_func is None:
                return
            self.as_ = s.as_
            rv = self.init_view_func(view, s.options)
            del self.as_
            return rv

        attrs = dict(kwargs, name=self.name, execute=execute,
            _execute=_execute, init_view=init_view)

        if hasattr(func, "actions"):
            # Override the methods property of the ActionFunction action
            # to return properties needed by the WithActionsDecorator actions.
            def methods(s):
                return func.actions.methods
            attrs.update(methods=property(methods))

        self.action = type(self.name, (Action,), attrs)

    def __call__(self, *args, **kwargs):
        with ensure_context():
            self.as_ = None
            rv = self.func(*args, **kwargs)
        return rv

    def init_view(self, f):
        self.unbound_init_view_func = f
        self.init_view_func = self._wrap_func(f)
        return f


class ActionProxy(Action):
    def __init__(self, action):
        self.action = action

    @property
    def name(self):
        return self.action.name

    @property
    def group(self):
        return self.action.group

    @property
    def methods(self):
        return self.action.methods

    @property
    def as_(self):
        return self.action.as_

    @property
    def requires(self):
        return self.action.requires

    @property
    def defaults(self):
        return self.action.defaults

    @property
    def default_option(self):
        return self.action.default_option

    @property
    def options(self):
        return self.action.options

    def init_view(self, *args, **kwargs):
        return self.action.init_view(*args, **kwargs)

    def is_valid(self, *args, **kwargs):
        return self.action.is_valid(*args, **kwargs)

    def execute(self, *args, **kwargs):
        return self.action.execute(*args, **kwargs)

    def _execute(self, *args, **kwargs):
        return self.action._execute(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)

    def __repr__(self):
        return '<%s(%s)>' % (self.__class__.__name__, repr(self.action))


from flask import current_app
from .actions import (Action, ActionFunction, ActionList, ensure_context,\
                     execute_actions, load_actions, current_context,\
                     ContextExitException, ReturnValueException)
from .utils import ClassBoundDecoratorMixin
from werkzeug.urls import url_join
import functools


def hook(name=None, *args, **kwargs):
    """Decorator to register the function as a hook
    """
    def decorator(f):
        if not hasattr(f, "hooks"):
            f.hooks = []
        f.hooks.append((name or f.__name__, args, kwargs))
        return f
    return decorator


def register_hooks(func, hooks, obj):
    """Register func on obj via hooks.
    Hooks should be a tuple of (name, args, kwargs) where
    name is a method name of obj. If args or kwargs are not empty,
    the method will be called first and expect a new function as return.
    """
    for name, args, kwargs in hooks:
        hook = getattr(obj, name)
        force_call = kwargs.pop("_force_call", False)
        if force_call or len(args) > 0 or len(kwargs) > 0:
            hook = hook(*args, **kwargs)
        hook(func)


def action(*args, **kwargs):
    """Transforms functions or class methods into actions.
    Optionnaly, you can define a function to be used as the view initializer:

        @action()
        def my_action():
            pass

        @my_action.init_view
        def my_action_init_view(view, options):
            pass
    """
    def decorator(f):
        return ActionFunction(f, *args, **kwargs)
    return decorator


class WithActionsDecorator(ClassBoundDecoratorMixin):
    """Decorator to execute actions before/after a function
    """
    def __init__(self, func, actions=None):
        self.unbound_func = func
        self.func = self._wrap_func(func)
        self.__name__ = func.__name__
        self.actions = ActionList(actions or [])

    def __call__(self, *args, **kwargs):
        rv = None
        with ensure_context(), current_context.clone(**kwargs):
            ctx_exit = None
            try:
                execute_actions(self.actions, ("before", None),
                    catch_context_exit=False)
                rv = self.func(*args, **kwargs)
            except ReturnValueException as e:
                rv = e.value
            except ContextExitException as e:
                ctx_exit = e
                current_context["response"] = e.result
            execute_actions(self.actions, ("after",),
                catch_context_exit=False)
            if ctx_exit:
                raise ctx_exit
        return rv


def with_actions(actions_or_group_name, actions=None):
    """Executes the list of actions before/after the function
    Actions should be a list where items are action names as
    strings or a dict. See frasco.actions.loaders.load_action().
    """
    group = None
    if isinstance(actions_or_group_name, str):
        group = actions_or_group_name
    else:
        actions = actions_or_group_name
    def decorator(f):
        if isinstance(f, WithActionsDecorator):
            dec = f
        else:
            dec = WithActionsDecorator(f)
        dec.actions.extend(load_actions(actions, group=group))
        return dec
    return decorator


def expose(rule, **options):
    """Decorator to add an url rule to a function
    """
    def decorator(f):
        if not hasattr(f, "urls"):
            f.urls = []
        if isinstance(rule, (list, tuple)):
            f.urls.extend(rule)
        else:
            f.urls.append((rule, options))
        return f
    return decorator




from jinja2 import Markup, evalcontextfilter
import re
import ago


__all__ = ('html_tag', 'html_attributes', 'nl2br', 'timeago', 'plural')


def html_tag(tagname, attrs=None, **kwargs):
    html_attrs = html_attributes(attrs, **kwargs)
    return Markup("<%s%s>" % (tagname, " " + html_attrs if html_attrs else ""))


def html_attributes(attrs=None, **kwargs):
    attrs = dict(attrs or {})
    attrs.update(kwargs)
    html = []
    for k, v in attrs.items():
        if v is None:
            continue
        if k.endswith('_'):
            k = k[:-1]
        if isinstance(v, bool):
            v = k if v else ""
        k = k.replace("_", "-")
        html.append('%s="%s"' % (k, str(v).strip()))
    return Markup(" ".join(html))


_paragraph_re = re.compile(r'(?:\r\n|\r(?!\n)|\n){2,}')

@evalcontextfilter
def nl2br(eval_ctx, value):
    result = u'\n\n'.join(u'<p>%s</p>' % p.replace(u'\n', Markup('<br>\n'))
                                         for p in _paragraph_re.split(value))
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def timeago(dt, **args):
    return ago.human(dt, **args)


def plural(singular, plural, num):
    if num > 1:
        return plural
    return singular


from ..expression import Parser as ExpressionParser
from .helpers import ActionList, ActionsAction
from .core import Action, ActionFunction
from .decorators import action_decorators
from .registry import action_resolver


def load_actions(spec, group=None, expr_parser=None):
    """Each item can be an action name as a string or a dict. When using a dict,
    one key/item pair must be the action name and its options and the rest action
    decorator names and their options.
    Example:
        load_actions(["login_required", {"flash": {"message": "hello world", "label": "warning"}}])
    """
    if expr_parser is None:
        expr_parser = ExpressionParser()
    actions = ActionList()

    for name in spec:
        options = {}
        as_ = None
        decorators = []

        if isinstance(name, dict):
            actionspec = dict(name)
            as_ = actionspec.pop("as", None)
            for dec, dec_cls in action_decorators:
                if dec in actionspec:
                    decorators.append((dec_cls, expr_parser.compile(actionspec.pop(dec))))
            name, options = actionspec.popitem()
            if options:
                options = expr_parser.compile(options)

        if isinstance(name, Action):
            action = name
        elif isinstance(name, ActionFunction):
            action = name.action
        else:
            action = action_resolver.resolve_or_delayed(name, options, group, as_)

        for dec_cls, arg in decorators:
            action = dec_cls(action, arg)

        actions.append(action)

    return actions


def load_grouped_actions(spec, default_group=None, key_prefix="actions", pop_keys=False, expr_parser=None):
    """Instanciates actions from a dict. Will look for a key name key_prefix and
    for key starting with key_prefix followed by a dot and a group name. A group
    name can be any string and will can be used later to filter actions.
    Values associated to these keys should be lists that will be loaded using load_actions()
    """
    actions = ActionList()
    if expr_parser is None:
        expr_parser = ExpressionParser()
    for key in spec.keys():
        if key != key_prefix and not key.startswith(key_prefix + "."):
            continue
        group = default_group
        if "." in key:
            (_, group) = key.split(".")
        actions.extend(load_actions(spec[key], group, expr_parser))
        if pop_keys:
            spec.pop(key)
    return actions


def create_action_from_dict(name, spec, base_class=ActionsAction, metaclass=type, pop_keys=False):
    """Creates an action class based on a dict loaded using load_grouped_actions()
    """
    actions = load_grouped_actions(spec, pop_keys=pop_keys)
    attrs = {"actions": actions, "name": name}
    if "as" in spec:
        attrs["as_"] = spec["as"]
        if pop_keys:
            del spec["as"]
    for k in ("requires", "methods", "defaults", "default_option"):
        if k in spec:
            attrs[k] = spec[k]
            if pop_keys:
                del spec[k]
    return metaclass(name, (base_class,), attrs)


from flask import *
from werkzeug.utils import cached_property

from .actions import (Action, load_actions, execute_action, execute_actions, current_context,\
                     ensure_context, Context, OptionMissingError, InvalidOptionError, Context,\
                     ActionRegistry, ContextExitException, resolve_action, pass_context_var)
from .features import Feature, pass_feature, feature_proxy
from .views import View, MethodView, ActionsView, Blueprint, as_view
from .services import Service, ServiceError, service_proxy, pass_service, service_mixin
from .request_params import (request_param, request_params, partial_request_param, request_param_loader,\
                     partial_request_param_loader, auto_inject_request_params)
from .marshaller import marshal_with, marshal_many_with, marshal_dict_with, marshaller_context, marshalling_context
from .decorators import hook, action, with_actions, expose
from .signals import signal, listens_to, create_action_listener
from .templating import render_layout, html_tag, html_attributes
from .app import Frasco
from .declarative import (DeclarativeFrasco, DeclarativeFrascoFactory, DeclarativeBlueprint,\
                          DeclarativeView, create_app_from_cwd, decl_app)
from .trans import (set_translation_callbacks, translatable, translate, ntranslate, lazy_translate, _,\
                    format_datetime, format_date, format_time)
from .commands import command, shell_exec
from .utils import (url_for_static, populate_obj, wrap_in_markup, AttrDict, slugify, import_string,\
                   copy_extra_feature_options)



from .core import ActionFunction, ActionProxy
from contextlib import contextmanager


class ActionNotFoundError(Exception):
    pass


class ActionResolver(object):
    def __init__(self):
        self.stack = []

    def push(self, registry, bubbleup=True):
        self.stack.append((registry, bubbleup))

    def pop(self):
        self.stack.pop()

    def resolve(self, name):
        for registry, bubbleup in reversed(self.stack):
            try:
                return registry.resolve(name)
            except ActionNotFoundError:
                pass
            if not bubbleup:
                break
        raise ActionNotFoundError("Action '%s' does not exist" % name)

    def resolve_or_delayed(self, name, *args, **kwargs):
        try:
            action_class = self.resolve(name)
            return action_class(*args, **kwargs)
        except ActionNotFoundError:
            return DelayedInstanciationAction(name, *args, **kwargs)


action_resolver = ActionResolver()


def resolve_action(name):
    return action_resolver.resolve(name)


class ActionRegistry(object):
    """Register and access action classes. If a package name
    is specified when registering an action, the action will be
    available through its short name or through package_name.action_name
    """
    def __init__(self, actions=None, package=None):
        self.actions = {}
        self.aliases = {}
        if actions:
            self.register_many(actions, package)

    def register(self, action, package=None):
        if isinstance(action, ActionRegistry):
            return self.merge(action, package)
        if isinstance(action, ActionFunction):
            action = action.action
        fullname = action.name
        if package is not None:
            fullname = package + "." + fullname
        self.actions[fullname] = action
        self.aliases[action.name] = fullname

    def register_many(self, actions, package=None):
        if isinstance(actions, ActionRegistry):
            return self.merge(actions, package)
        for action in actions:
            self.register(action, package)

    def merge(self, registry, package=None):
        for action in registry:
            self.register(action, package)

    def action(self, *args, **kwargs):
        def decorator(f):
            af = ActionFunction(f, *args, **kwargs)
            self.register(af.action)
            return af
        return decorator

    def resolve(self, name):
        if name in self.aliases:
            return self.actions[self.aliases[name]]
        if name in self.actions:
            return self.actions[name]
        raise ActionNotFoundError("Action '%s' does not exist" % name)

    def as_only_resolver(self):
        action_resolver.push(self, False)
        try:
            yield self
        finally:
            action_resolver.pop()

    def __getitem__(self, name):
        return self.resolve(name)

    def __getattr__(self, name):
        return self.resolve(name)

    def __contains__(self, name):
        return name in self.aliases or name in self.actions

    def __iter__(self):
        return self.actions.itervalues()

    def __enter__(self):
        action_resolver.push(self)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        action_resolver.pop()
        return False


class DelayedInstanciationAction(ActionProxy):
    """Wrap action constructor args for later initialization
    """
    def __init__(self, name, *ctorargs, **ctorkwargs):
        self._name = name
        self.ctorargs = ctorargs
        self.ctorkwargs = ctorkwargs
        self._action = None

    @property
    def name(self):
        return self._name
    
    @property
    def action(self):
        if not self._action:
            action_class = action_resolver.resolve(self.name)
            self._action = action_class(*self.ctorargs, **self.ctorkwargs)
        return self._action

    def __repr__(self):
        return '<%s(%s, %s, %s)>' % (self.__class__.__name__, self._name, self.ctorargs, self.ctorkwargs)


from .actions import (ActionFunction, ActionList, current_context, ContextExitException,\
                     ensure_context)
from .decorators import WithActionsDecorator, register_hooks, expose
from .views import Blueprint, as_view, ActionsView, full_exec_request_actions
from .marshaller import marshaller_context, marshal
from .request_params import auto_inject_request_params
from .utils import logger, preprend_base_url_to_urls
from flask import json, request, current_app, jsonify, abort, make_response
from werkzeug.local import LocalProxy
from werkzeug.exceptions import HTTPException
from werkzeug.wrappers import Response
import inspect
import functools
import copy


class ServiceActionsView(ActionsView):
    def __init__(self, *args, **kwargs):
        self.marshaller = kwargs.pop('marshaller', None)
        super(ServiceActionsView, self).__init__(*args, **kwargs)

    """View class for service views created using the ServiceLoader
    """
    def dispatch_request(self, *args, **kwargs):
        try:
            rv = super(ServiceActionsView, self).dispatch_request(*args, **kwargs)
        except ServiceError as e:
            return make_response(json.dumps({"error": e.message}), e.http_code)
        if isinstance(rv, Response):
            return rv
        # child actions can use "return" to return a value which
        # can be something else than a proper Response instance.
        # In this case we encode the return value to json
        return marshal(rv, self.marshaller)

    def _auto_render_func(self):
        # the _auto_render_func() func is only called if the actions do not
        # exit the context, thus we return an empty json response
        return jsonify()


def get_current_app_service(name):
    return current_app.services[name]


_services_proxies = {}

def service_proxy(name):
    if name not in _services_proxies:
        _services_proxies[name] = LocalProxy(lambda: get_current_app_service(name))
    return _services_proxies[name]


def pass_service(*names):
    """Injects a service instance into the kwargs
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            for name in names:
                kwargs[name] = service_proxy(name)
            return f(*args, **kwargs)
        return wrapper
    return decorator


class ServiceError(Exception):
    def __init__(self, message, http_code=500):
        self.message = message
        self.http_code = http_code


class ServiceMeta(type):
    def __init__(cls, name, bases, attrs):
        type.__init__(cls, name, bases, attrs)
        cls.current = service_proxy(attrs.get('name') or name)


def json_service_marshaller(rv):
    return json.dumps(rv), {'Content-Type': current_app.config.get(
        'SERVICES_CONTENT_TYPE', 'application/json;charset=UTF-8')}


class Service(object):
    name = None
    view_class = ServiceActionsView
    blueprint_class = Blueprint
    url_prefix = None
    marshaller = staticmethod(json_service_marshaller)
    __metaclass__ = ServiceMeta

    def __init__(self):
        self.name = self.name or self.__class__.__name__
        self.actions = ActionList()
        self.views = []
        self.hooks = []
        for attr in dir(self):
            attr = getattr(self, attr)
            if isinstance(attr, ActionFunction):
                self.actions.append(attr.action)
            if hasattr(attr, 'urls'):
                self._expose(attr, attr.urls)
            if hasattr(attr, 'hooks'):
                self.hooks.append(attr)

    def _expose(self, func, urls):
        name = func.__name__
        if isinstance(func, ActionFunction):
            name = func.unbound_func.__name__
            func = func.unbound_func if isinstance(func.unbound_func, WithActionsDecorator) else func.func
        if self.url_prefix:
            preprend_base_url_to_urls(self.url_prefix, urls)
        self.views.append(as_view(name=name,
                                  url_rules=urls,
                                  view_class=self.view_class,
                                  marshaller=self.marshaller)(func))

    @property
    def exposed(self):
        return len(self.views) > 0 or len(self.hooks) > 0

    @property
    def blueprint_name(self):
        return "%s_service" % self.name

    def as_blueprint(self):
        bp = self.blueprint_class(self.blueprint_name, self.__class__.__module__)
        for view in self.views:
            bp.add_view(view)
        for attr in self.hooks:
            register_hooks(attr, attr.hooks, bp)
        return bp


def service_mixin(cls, decorator, url_prefix=None):
    attrs = {}
    for name in dir(cls):
        attr = getattr(cls, name)
        if hasattr(attr, 'urls'):
            urls = copy.deepcopy(attr.urls)
            if url_prefix:
                preprend_base_url_to_urls(url_prefix, urls)
            decorated = decorator(attr)
            decorated.urls = urls
            attrs[attr.__name__] = decorated
    return type(cls.__name__, (cls,), attrs)



import yaml
import os
import inspect
from ..views import Blueprint, View, ActionsView
from ..actions import load_grouped_actions
from ..utils import import_string, parse_yaml_frontmatter


def create_blueprint_from_dict(name, spec, import_name="__main__", blueprint_class=Blueprint, **kwargs):
    spec = dict(spec)
    actions = []
    if "actions" in spec:
        actions = load_grouped_actions(spec, pop_keys=True)

    views = []
    if "views" in spec:
        for name, viewspec in spec.pop("views").iteritems():
            views.append(create_view_from_dict(name, viewspec))
    
    bp = blueprint_class(name, import_name, **dict(spec, **kwargs))
    for a in actions:
        bp.add_request_action(a)
    for v in views:
        bp.add_view(v)
    return bp


def create_blueprint_from_file(filename, import_name='__main__', name=None, blueprint_class=Blueprint, **kwargs):
    with open(filename) as f:
        spec = yaml.load(f.read())
    if "name" in spec:
        name = spec.pop("name")
    elif not name:
        name, _ = os.path.splitext(os.path.basename(path))
    if "extends" in spec:
        blueprint_class = import_string(spec.pop("extends"))
    kwargs.setdefault('root_path', os.path.dirname(filename))
    return create_blueprint_from_dict(name, spec, import_name, blueprint_class, **kwargs)


def create_blueprint_views_from_path(bp, path):
    for f in os.listdir(path):
        filename = os.path.join(path, f)
        if f.startswith("_") or os.path.isdir(filename):
            continue
        view_name = os.path.splitext(f)[0]
        template = os.path.join(bp.name, f)
        try:
            view = create_view_from_file(filename, name=view_name, source_template=template)
            bp.add_view(view)
        except DeclarativeViewError:
            pass


def create_blueprint_from_path(path, import_name='__main__', name=None, blueprint_class=Blueprint):
    if not name:
        name, _ = os.path.splitext(os.path.basename(os.path.realpath(path)))
    init_file = os.path.join(path, "__init__.yml")
    kwargs = dict(root_path=path, template_folder='.', template_prefix=name)
    if os.path.exists(init_file):
        bp = create_blueprint_from_file(init_file, import_name, name,
            blueprint_class, **kwargs)
    else:
        bp = blueprint_class(name, import_name, **kwargs)
    create_blueprint_views_from_path(bp, path)
    return bp


def create_view_from_dict(name, spec, template=None, cls=ActionsView):
    """Creates a view from an spec dict (typically, the YAML front-matter).
    """
    kwargs = dict(spec)
    if template is not None:
        kwargs.setdefault("template", template)
    actions = load_grouped_actions(kwargs, pop_keys=True)
    view = cls(name=name, **kwargs)
    if isinstance(view, ActionsView):
        view.actions.extend(actions)
    return view


def create_view_from_file(filename, template=None, name=None, source_template=None, view_class=ActionsView):
    with open(filename) as f:
        source = f.read()

    if not source.startswith("---\n"):
        raise DeclarativeViewError("Declarative views must start with a YAML front-matter")

    basename = os.path.basename(filename)
    spec, source = parse_yaml_frontmatter(source)
    if source is not None and not template:
        template = source_template or basename

    if "name" in spec:
        name = spec.pop("name")
    elif not name:
        name = basename.split(".")[0]
    if "extends" in spec:
        view_class = import_string(spec.pop("extends"))

    return create_view_from_dict(name, spec, template, view_class)


class DeclarativeBlueprint(Blueprint):
    def __new__(cls, path_or_filename, import_name='__main__', name=None):
        if os.path.isdir(path_or_filename):
            return create_blueprint_from_path(path_or_filename, import_name, name)
        return create_blueprint_from_file(path_or_filename, import_name, name)


class DeclarativeView(ActionsView):
    def __new__(cls, filename, template=None, name=None, source_template=None):
        return create_view_from_file(filename, template, name, source_template)


class DeclarativeViewError(Exception):
    pass


from setuptools import setup, find_packages


setup(
    name='frasco',
    version='0.6.12',
    url='http://github.com/frascoweb/frasco',
    license='MIT',
    author='Maxime Bouroumeau-Fuseau',
    author_email='maxime.bouroumeau@gmail.com',
    description='Build web applications without coding',
    packages=find_packages(),
    package_data={
        'frasco': ['templating/*.html'],
    },
    zip_safe=False,
    platforms='any',
    install_requires=[
        'flask>=0.11',
        'honcho>=0.6',
        'PyYAML>=3.11',
        'blinker>=1.3',
        'jinja-macro-tags',
        'jinja-layout',
        'python-slugify',
        'ago',
        'simplejson',
        'speaklater',
        'requests>=2.3.0'
    ],
    entry_points='''
        [console_scripts]
        frasco=frasco.cli:main
    '''
)



from .context import ContextExitException, TriggerActionGroupException, current_context, ensure_context
from .core import Action
from ..utils import logger


class ActionList(list):
    """A list of action. Makes it easy to filter
    """
    @property
    def groups(self):
        groups = set()
        for action in self:
            groups.add(action.group)
        return groups

    @property
    def methods(self):
        methods = set()
        for action in self:
            if not action.group and action.methods:
                methods.update(action.methods)
            elif action.group in ("get", "post", "put", "delete", "head", "options"):
                methods.add(action.group.upper())
        if not methods:
            methods.add("GET")
        return methods

    @property
    def requirements(self):
        reqs = set()
        for action in self:
            reqs.update(action.requires)
        return reqs

    def init_view(self, view):
        for action in self:
            action.init_view(view)

    def has_action(self, action_name):
        for action in self:
            if action.name == action_name:
                return True
        return False

    def filter_by_action(self, *action_names):
        for action in self:
            if action.name in action_names:
                yield action

    def filter_by_group(self, *groups):
        for action in self:
            if action.group in groups:
                yield action

    def execute(self, **kwargs):
        return execute_actions(self, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.execute(*args, **kwargs)


class InvalidActionError(Exception):
    pass


def execute_action(action, catch_context_exit=True):
    with ensure_context():
        try:
            if not action.is_valid(current_context):
                logger.debug("Skipping invalid action %s" % action.name)
                raise InvalidActionError
            current_context.executed_actions.add(action.name)
            logger.debug("Executing action %s" % action.name)
            return action.execute()
        except ContextExitException as e:
            if not catch_context_exit:
                raise
            return e.result


def execute_actions(actions, limit_groups=None, catch_context_exit=True, trigger_error_group_on_exc=True):
    if limit_groups and not isinstance(limit_groups, (list, tuple)):
        limit_groups = [limit_groups]
    with ensure_context():
        try:
            if limit_groups and len(limit_groups) > 1:
                for group in limit_groups:
                    execute_actions(actions, (group,), False)
            else:
                for action in actions:
                    if limit_groups and action.group not in limit_groups:
                        continue
                    try:
                        execute_action(action, False)
                    except InvalidActionError:
                        continue
                    except TriggerActionGroupException as e:
                        logger.debug("Triggering action group %s" % e.group)
                        execute_actions(actions, (e.group,), False)
                    except Exception as e:
                        if trigger_error_group_on_exc:
                            execute_actions(actions, ("%s_error" % action.name,), False)
                        raise

        except ContextExitException as e:
            rv = e.result
            if e.trigger_action_group is not None:
                logger.debug("Triggering action group %s after context exit" % e.trigger_action_group)
                rv = execute_actions(actions, (e.trigger_action_group,)) or rv
            if not catch_context_exit:
                raise ContextExitException(rv)
            return rv


class ReturnValueException(Exception):
    def __init__(self, value):
        super(ReturnValueException, self).__init__()
        self.value = value


class ActionsAction(Action):
    """Represents an action that executes a list of actions when executed.
    Child actions can raise a ReturnValueException which will hold the return
    value of the action.
    """
    actions = None

    def __init__(self, *args, **kwargs):
        actions = kwargs.pop('actions', None)
        super(ActionsAction, self).__init__(*args, **kwargs)
        self.actions = ActionList(self.actions or [])
        if actions:
            self.actions.extend(actions)

    @property
    def methods(self):
        return self.actions.methods

    def init_view(self, view):
        for action in self.actions:
            action.init_view(view)

    def _execute(self, options):
        with current_context.clone(**options):
            try:
                execute_actions(self.actions, (None,), catch_context_exit=False)
            except ReturnValueException as e:
                return e.value

    def __call__(self, **kwargs):
        return self._execute(kwargs)


import flask
import flask.views
from flask.helpers import locked_cached_property
from jinja2 import FileSystemLoader, PrefixLoader
import yaml
from .utils import AttrDict, logger
from .actions import (ActionList, execute_actions, ContextExitException, current_context,\
                     execute_action, load_actions, ReturnValueException)
from .decorators import hook, register_hooks, expose, WithActionsDecorator
import os
import inspect


def exec_before_request_actions(actions, **kwargs):
    """Execute actions in the "before" and "before_METHOD" groups
    """
    groups = ("before", "before_" + flask.request.method.lower())
    return execute_actions(actions, limit_groups=groups, **kwargs)


def exec_request_actions(actions, **kwargs):
    """Execute actions in the None and METHOD groups
    """
    groups = (None, flask.request.method.lower())
    return execute_actions(actions, limit_groups=groups, **kwargs)


def exec_after_request_actions(actions, response, **kwargs):
    """Executes actions of the "after" and "after_METHOD" groups.
    A "response" var will be injected in the current context.
    """
    current_context["response"] = response
    groups = ("after_" + flask.request.method.lower(), "after")
    try:
        rv = execute_actions(actions, limit_groups=groups, **kwargs)
    except ReturnValueException as e:
        rv = e.value
    if rv:
        return rv
    return response


def full_exec_request_actions(actions, func=None, render_func=None):
    """Full process to execute before, during and after actions.
    If func is specified, it will be called after exec_request_actions()
    unless a ContextExitException was raised. If render_func is specified,
    it will be called after exec_request_actions() only if there is no
    response. exec_after_request_actions() is always called.
    """
    response = None
    try:
        exec_before_request_actions(actions, catch_context_exit=False)
        exec_request_actions(actions, catch_context_exit=False)
        if func:
            response = func()
    except ContextExitException as e:
        response = e.result
    except ReturnValueException as e:
        response = e.value
    if render_func and response is None:
        response = render_func()
    return exec_after_request_actions(actions, response)


class RegistrableViewMixin(object):
    """A mixin to add the register() method to views. This method is
    called when a blueprint is registered
    """
    name = None
    # If url_rules is None, an url rule will be created using url and methods
    url = None
    methods = None
    url_rules = None

    def __init__(self, name=None, url=None, methods=None, url_rules=None):
        if name is not None:
            self.name = name
        elif self.name is None:
            self.name = self.__class__.__name__
        if url_rules is not None:
            self.url_rules = url_rules
        elif self.url_rules is None:
            url = url or self.url or "/" + self.name
            if isinstance(url, str):
                url = [url]
            methods = methods or self.methods
            self.url_rules = [(u, {"methods": methods}) for u in url]

    def register(self, target):
        """Registers url_rules on the blueprint
        """
        for rule, options in self.url_rules:
            target.add_url_rule(rule, self.name, self.dispatch_request, **options)


class View(RegistrableViewMixin, flask.views.View):
    """Same as flask.views.View but with RegistrableViewMixin"""


class MethodView(RegistrableViewMixin, flask.views.MethodView):
    """Same as flask.views.MethodView but with RegistrableViewMixin"""


class ActionsView(View):
    """Executes a list of actions and renders a template.
    """
    actions = None
    template = None
    auto_render = True
    func = None

    def __init__(self, name=None, actions=None, func=None, template=None, auto_render=None,
                 self_var=None, **kwargs):
        super(ActionsView, self).__init__(name=name, **kwargs)
        self.actions = ActionList(actions or self.actions or [])
        self.func = func or self.func
        if template is not None:
            self.template = template
        if auto_render is not None:
            self.auto_render = auto_render
        self.self_var = self_var or self

    def add_action(self, action):
        self.actions.append(action)

    def register(self, target):
        self.actions.init_view(self)
        default_methods = self.actions.methods
        for rule, options in self.url_rules:
            if options.get("methods") is None:
                options["methods"] = default_methods
            logger.info("Registering url rule '%s' with url '%s' on methods %s" % (self.name, rule, options["methods"]))
            target.add_url_rule(rule, self.name, self.dispatch_request, **options)

    def dispatch_request(self, *args, **kwargs):
        logger.debug("Dispatching view %s" % self.name)
        current_context.vars.self = self.self_var
        r = self.execute_actions(*args, **kwargs)
        current_context.vars.pop('self', None)
        return r

    def execute_actions(self, *args, **kwargs):
        func = None
        if self.func:
            func = lambda: self.func(*args, **kwargs)
        return full_exec_request_actions(self.actions, func=func, render_func=self._auto_render_func)

    def _auto_render_func(self):
        if self.auto_render:
            current_context.vars.pop('self', None)
            return self.render(**current_context.vars)

    def render(self, **kwargs):
        if self.template is not None:
            logger.debug("Rendering template %s" % self.template)
            return flask.render_template(self.template, **kwargs)


def as_view(url=None, methods=None, view_class=ActionsView, name=None, url_rules=None, **kwargs):
    """Decorator to transform a function into a view class. Be warned that this will replace
    the function with the view class.
    """
    def decorator(f):
        if url is not None:
            f = expose(url, methods=methods)(f)

        clsdict = {"name": name or f.__name__,
                   "actions": getattr(f, "actions", None),
                   "url_rules": url_rules or getattr(f, "urls", None)}

        if isinstance(f, WithActionsDecorator):
            f = f.func
        clsdict['func'] = f

        def constructor(self, **ctorkwargs):
            for k, v in kwargs.items():
                if k not in ctorkwargs or ctorkwargs[k] is None:
                    ctorkwargs[k] = v
            view_class.__init__(self, func=f, **ctorkwargs)

        clsdict["__init__"] = constructor
        return type(f.__name__, (view_class,), clsdict)

    return decorator


class ViewContainerMixin(object):
    """Mixin to add view registration on a class.
    """
    view_class = ActionsView

    def add_view(self, view):
        rv = view
        if inspect.isclass(view):
            view = view(self_var=self)
        self.views[view.name] = view
        return rv

    def view(self, *args, **kwargs):
        """Decorator to automatically apply as_view decorator and register it.
        """
        def decorator(f):
            kwargs.setdefault("view_class", self.view_class)
            return self.add_view(as_view(*args, **kwargs)(f))
        return decorator

    def add_action_view(self, name, url, actions, **kwargs):
        """Creates an ActionsView instance and registers it.
        """
        view = ActionsView(name, url=url, self_var=self, **kwargs)
        if isinstance(actions, dict):
            for group, actions in actions.iteritems():
                view.actions.extend(load_actions(actions, group=group or None))
        else:
            view.actions.extend(load_actions(actions))
        self.add_view(view)
        return view


class Blueprint(ViewContainerMixin, flask.Blueprint):
    """Blueprint class which adds support for actions before/after request
    and to add views
    """
    name = None
    url_prefix = None
    subdomain = None

    def __init__(self, name=None, import_name=None, *args, **kwargs):
        if (name is None or import_name is None) and self.__class__ == Blueprint:
            raise AttributeError("Missing name and import_name parameter. Omitting them requires to subclass Blueprint")
        self.template_prefix = kwargs.pop('template_prefix', None)
        kwargs.setdefault('url_prefix', self.url_prefix)
        kwargs.setdefault('subdomain', self.subdomain)
        super(Blueprint, self).__init__(name or self.name or self.__class__.__name__,
            import_name or self.__class__.__module__, *args, **kwargs)
        self.request_actions = ActionList()
        self.views = {}
        self.register_defaults()

    def register_defaults(self):
        for attr in dir(self):
            attr = getattr(self, attr)
            if hasattr(attr, "urls"):
                self.view()(attr)
            elif hasattr(attr, 'hooks'):
                register_hooks(attr, attr.hooks, self)

    def add_request_action(self, action):
        self.request_actions.append(action)

    @locked_cached_property
    def jinja_loader(self):
        if self.template_folder is None:
            return
        loader = FileSystemLoader(os.path.join(self.root_path, self.template_folder))
        if self.template_prefix:
            loader = PrefixLoader(dict([(self.template_prefix, loader)]))
        return loader

    @hook('record')
    def register_views(self, state):
        logger.info("Registering blueprint '%s' (url_prefix=%s, subdomain=%s)" % (self.name, self.url_prefix, self.subdomain))
        for view in self.views.values():
            view.register(self)

    @hook('before_request')
    def exec_before_request_actions(self, *args, **kwargs):
        try:
            exec_before_request_actions(self.request_actions, catch_context_exit=False)
            exec_request_actions(self.request_actions, catch_context_exit=False)
        except ContextExitException as e:
            return e.result

    @hook('after_request')
    def exec_after_request_actions(self, response):
        return exec_after_request_actions(self.request_actions, response)



from ..views import Blueprint, ActionsView, View
from ..actions import load_grouped_actions, create_action_from_dict
from ..features import FeaturesLoader
from ..services import Service, ServiceActionsView
from ..utils import import_string, find_classes_in_module, RequirementMissingError, deep_update_dict
from .views import create_blueprint_from_file, create_view_from_file, DeclarativeViewError
from collections import OrderedDict
import os
import yaml
import inspect
import re


class FileLoader(object):
    """Generic file loader. Will iterate over directories, load files and
    python modules
    """
    # Filter file by extensions (must be prefixed with a dot).
    # Set to None to not filter
    allowed_extensions = None
    # Whether to exclude files starting with an underscore
    exclude_file_starting_with_dash = True
    # Base python path
    pypath = ""
    # Whether to load dirs recursively
    recursive = False

    def __init__(self, path, pypath=None, recursive=None, exclude_file_starting_with_dash=None):
        self.path = path
        if pypath is not None:
            self.pypath = pypath
        if recursive is not None:
            self.recursive = recursive
        if exclude_file_starting_with_dash:
            self.exclude_file_starting_with_dash = exclude_file_starting_with_dash

    def load(self, app):
        return self.load_dir(app, self.path, "", self.pypath)

    def load_dir(self, app, path, relpath, pypath):
        objects = []
        if not os.path.exists(path):
            return objects
        for filename in os.listdir(path):
            if self.exclude_file_starting_with_dash and filename.startswith("_"):
                continue

            pathname = os.path.join(path, filename)
            (pyname, ext) = os.path.splitext(filename)
            filepypath = (pypath + "." + pyname).lstrip(".")
            filerelpath = os.path.join(relpath, filename)

            if os.path.isdir(pathname) and self.recursive:
                objects.extend(self.load_dir(app, pathname, filerelpath, filepypath))
            elif self.allowed_extensions is None or ext in self.allowed_extensions:
                objects.append(self.load_file(app, pathname, filerelpath, filepypath))

        return filter(lambda c: c is not None, objects)

    def load_file(self, app, filename, relpath, pypath):
        """Loads a file from a directory. If returns None, it will be ignored
        """
        return None


class ActionsLoader(FileLoader):
    """Loads Action classes from the given path.
    Actions can be loaded from python modules or from YAML files.
    In the latter case, a sublass of ActionsAction will be created using
    create_action_from_dict() to create the list of actions
    """
    allowed_extensions = (".yml",)

    def load_file(self, app, pathname, relpath, pypath):
        with open(pathname) as f:
            spec = yaml.load(f.read())
        return create_action_from_dict(pypath.split(".")[-1], spec)


class ViewsLoader(FileLoader):
    """Loads blueprints from python modules found in path and from features.
    Creates views out of yml and html files if they start with a YAML front-matter
    and creates blueprints out of directories
    """
    allowed_extensions = set([".yml", ".html"])
    recursive = True
    default_pkg_name = None
    default_view_class = ActionsView
    extend_feature_blueprints = True

    def __init__(self, path, default_pkg_name=None, default_view_class=None, extend_feature_blueprints=None, **kwargs):
        super(ViewsLoader, self).__init__(path, **kwargs)
        if default_pkg_name is not None:
            self.default_pkg_name = default_pkg_name
        if default_view_class is not None:
            self.default_view_class = default_view_class
        if extend_feature_blueprints is not None:
            self.extend_feature_blueprints = extend_feature_blueprints

    def load(self, app):
        # Features can define a "view_files" property which must
        # be a list where items are tuple or a file extension
        # and a view class. The view class will be used for files
        # matching the extension
        self.view_class_files_map = []
        for f in app.features:
            if hasattr(f, 'view_files') and f.view_files is not None:
                for pattern, cls in f.view_files:
                    self.add_view_file_mapping(pattern, cls)

        return super(ViewsLoader, self).load(app)

    def add_view_file_mapping(self, pattern, cls):
        """Adds a mapping between a file and a view class.
        Pattern can be an extension in the form .EXT or a filename.
        """
        if isinstance(pattern, str):
            if not pattern.endswith("*"):
                _, ext = os.path.splitext(pattern)
                self.allowed_extensions.add(ext)
            pattern = re.compile("^" + re.escape(pattern).replace("\\*", ".+") + "$", re.I)
        self.view_class_files_map.append((pattern, cls))

    def load_dir(self, app, path, relpath, pypath):
        bpname = pypath[len(self.pypath):].lstrip(".").replace(".", "_")
        if not bpname:
            bpname = self.default_pkg_name

        bp = None
        blueprints = []
        # An __init__.yml file can be used to define
        # options of a blueprint
        init_file = os.path.join(path, '__init__.yml')
        if not os.path.exists(init_file):
            init_file = None
        if not bpname and init_file:
            raise DeclarativeViewError('__init__.yml cannot be used in the root of the views folder')
        if bpname and not init_file and self.extend_feature_blueprints:
            # If a feature-provided blueprint of the same name exists,
            # views found in this dir will be added to the existing
            # blueprint
            bp = self.find_feature_blueprint(app, bpname)

        if not bp and bpname:
            # If we are not using an existing blueprint, creates
            # a new blueprint instance
            if init_file:
                bp = create_blueprint_from_file(init_file, name=bpname)
            else:
                bp = Blueprint(bpname, "__main__")
            blueprints.append(bp)

        objects = super(ViewsLoader, self).load_dir(app, path, relpath, pypath)
        for obj in objects:
            if isinstance(obj, Blueprint) or not bp:
                blueprints.append(obj)
            elif isinstance(obj, View):
                bp.add_view(obj)

        return blueprints

    def find_feature_blueprint(self, app, name):
        for feature in app.features:
            if feature.blueprints:
                for bp in feature.blueprints:
                    if isinstance(bp, Blueprint) and bp.name == name:
                        return bp

    def load_file(self, app, pathname, relpath, pypath):
        """Loads a file and creates a View from it. Files are split
        between a YAML front-matter and the content (unless it is a .yml file).
        """
        try:
            view_class = self.get_file_view_cls(relpath)
            return create_view_from_file(pathname, source_template=relpath, view_class=view_class)
        except DeclarativeViewError:
            pass

    def get_file_view_cls(self, filename):
        """Returns the view class associated to a filename
        """
        if filename is None:
            return self.default_view_class
        for pattern, cls in self.view_class_files_map:
            if pattern.match(filename):
                return cls
        return self.default_view_class


class ServicesLoader(FileLoader):
    allowed_extensions = (".yml",)

    def load_file(self, app, pathname, relpath, pypath):
        with open(pathname) as f:
            config = yaml.load(f.read())

        name, _ = os.path.splitext(os.path.basename(pathname))
        service = Service()
        service.name = config.pop('name', name)
        service.url_prefix = config.pop("url_prefix", None)
        service.subdomain = config.pop("subdomain", None)

        for name, spec in config.iteritems():
            service.actions.append(self.create_action_from_spec(name, spec))
            service.views.append(self.create_view_from_spec(name, spec))

        return service

    def create_action_from_spec(self, name, spec):
        return create_action_from_dict(name, spec)

    def create_view_from_spec(self, name, spec):
        actions = load_grouped_actions(spec, pop_keys=True)
        for k in ("as", "requires", "defaults", "default_option"):
            spec.pop(k, None)
        return ServiceActionsView(name, actions=actions, **spec)